#######################################################################
# Copyright (c) 2019-present, Blosc Development Team <blosc@blosc.org>
# All rights reserved.
#
# SPDX-License-Identifier: BSD-3-Clause
#######################################################################

from __future__ import annotations

from dataclasses import dataclass
from typing import Any


@dataclass(frozen=True, slots=True)
class Ref:
    """A durable reference to a Blosc2 object.

    ``Ref`` can describe:

    - a persistent local Blosc2 object reopenable from ``urlpath``
    - a member inside a :class:`blosc2.DictStore`
    - a remote :class:`blosc2.C2Array`

    Instances can be created directly, from dictionaries via :meth:`from_dict`,
    or from supported objects via :meth:`from_object`. Use :meth:`open` to
    resolve the reference back into a live Blosc2 object.
    """

    kind: str
    urlpath: str | None = None
    key: str | None = None
    path: str | None = None
    urlbase: str | None = None

    def __post_init__(self) -> None:
        if self.kind == "urlpath":
            if not isinstance(self.urlpath, str):
                raise TypeError("Ref(kind='urlpath') requires a string 'urlpath'")
            if self.key is not None or self.path is not None or self.urlbase is not None:
                raise ValueError("Ref(kind='urlpath') only supports the 'urlpath' field")
            return
        if self.kind == "dictstore_key":
            if not isinstance(self.urlpath, str):
                raise TypeError("Ref(kind='dictstore_key') requires a string 'urlpath'")
            if not isinstance(self.key, str):
                raise TypeError("Ref(kind='dictstore_key') requires a string 'key'")
            if self.path is not None or self.urlbase is not None:
                raise ValueError("Ref(kind='dictstore_key') only supports 'urlpath' and 'key'")
            return
        if self.kind == "c2array":
            if not isinstance(self.path, str):
                raise TypeError("Ref(kind='c2array') requires a string 'path'")
            if self.urlbase is not None and not isinstance(self.urlbase, str):
                raise TypeError("Ref(kind='c2array') requires 'urlbase' to be a string or None")
            if self.urlpath is not None or self.key is not None:
                raise ValueError("Ref(kind='c2array') only supports 'path' and 'urlbase'")
            return
        raise ValueError(f"Unsupported Ref kind: {self.kind!r}")

    @classmethod
    def urlpath_ref(cls, urlpath: str) -> Ref:
        return cls(kind="urlpath", urlpath=urlpath)

    @classmethod
    def dictstore_key(cls, urlpath: str, key: str) -> Ref:
        return cls(kind="dictstore_key", urlpath=urlpath, key=key)

    @classmethod
    def c2array_ref(cls, path: str, urlbase: str | None = None) -> Ref:
        return cls(kind="c2array", path=path, urlbase=urlbase)

    @classmethod
    def from_dict(cls, payload: dict[str, Any]) -> Ref:
        if not isinstance(payload, dict):
            raise TypeError("Ref payload must be a mapping")
        version = payload.get("version")
        if version != 1:
            raise ValueError(f"Unsupported Ref payload version: {version!r}")
        return cls(
            kind=payload.get("kind"),
            urlpath=payload.get("urlpath"),
            key=payload.get("key"),
            path=payload.get("path"),
            urlbase=payload.get("urlbase"),
        )

    @classmethod
    def from_object(cls, obj: Any) -> Ref:
        import blosc2

        if isinstance(obj, blosc2.C2Array):
            return cls.c2array_ref(obj.path, obj.urlbase)
        if isinstance(obj, blosc2.Proxy):
            obj = obj._cache
        ref = getattr(obj, "_blosc2_ref", None)
        if isinstance(ref, cls):
            return ref
        if hasattr(obj, "schunk"):
            urlpath = obj.schunk.urlpath
            if urlpath is None:
                raise ValueError("Durable Blosc2 references require operands to be stored on disk/network")
            return cls.urlpath_ref(urlpath)
        raise TypeError("Durable Blosc2 references require NDArray, C2Array, or Proxy operands")

    def to_dict(self) -> dict[str, Any]:
        payload = {"kind": self.kind, "version": 1}
        if self.kind == "urlpath":
            payload["urlpath"] = self.urlpath
        elif self.kind == "dictstore_key":
            payload["urlpath"] = self.urlpath
            payload["key"] = self.key
        elif self.kind == "c2array":
            payload["path"] = self.path
            payload["urlbase"] = self.urlbase
        return payload

    def open(self):
        import blosc2

        if self.kind == "urlpath":
            # Structured refs are used to reopen operands for persisted recipes.
            # Read-only access avoids allocating unnecessary writable state.
            return blosc2.open(self.urlpath, mode="r")
        if self.kind == "dictstore_key":
            return blosc2.DictStore(self.urlpath, mode="r")[self.key]
        if self.kind == "c2array":
            return blosc2.C2Array(self.path, urlbase=self.urlbase)
        raise ValueError(f"Unsupported Ref kind: {self.kind!r}")



#######################################################################
# Copyright (c) 2019-present, Blosc Development Team <blosc@blosc.org>
# All rights reserved.
#
# SPDX-License-Identifier: BSD-3-Clause
#######################################################################

from __future__ import annotations

import os
import shutil
import tempfile
import warnings
import zipfile
from typing import TYPE_CHECKING, Any

import numpy as np

import blosc2
from blosc2.c2array import C2Array
from blosc2.embed_store import EmbedStore
from blosc2.schunk import SChunk, process_opened_object

if TYPE_CHECKING:
    from collections.abc import Iterator, Set


class DictStore:
    """
    Dictionary-like storage for compressed Blosc2 objects.

    Manages arrays in a directory or zip-file backed format.

    Supports the following types:

    - blosc2.NDArray: n-dimensional arrays. When persisted externally they
      are stored as .b2nd files.
    - blosc2.SChunk: super-chunks. When persisted externally they are stored
      as .b2f files.
    - blosc2.BatchArray: batched variable-length containers. When persisted
      externally they are stored as .b2b files.
    - blosc2.C2Array: columnar containers. These are always kept inside the
      embedded store (never externalized).
    - numpy.ndarray: converted to blosc2.NDArray on assignment.

    Parameters
    ----------
    localpath : str
        Local path for the directory or zip file. A ``.b2z`` suffix selects the
        compact zip-backed format. Existing directories, and new paths not
        ending in ``.b2z``, use Blosc2 directory format (B2DIR); a ``.b2d``
        suffix is recommended for these directory-backed stores. Existing files
        are treated as Blosc2 zip format (B2ZIP).
    mode : str, optional
        File mode ('r', 'w', 'a'). Default is 'a'.
    mmap_mode : str or None, optional
        Memory mapping mode for read access. For now, only ``"r"`` is supported,
        and only when ``mode="r"``. Default is None.
    tmpdir : str or None, optional
        Temporary directory to use when working with ".b2z" files. If None,
        a temporary directory is created in the same directory as the ".b2z"
        file, so that unpacked data stays on the same filesystem. Default is None.
    cparams : dict or None, optional
        Compression parameters for the internal embed store.
        If None, the default Blosc2 parameters are used.
    dparams : dict or None, optional
        Decompression parameters for the internal embed store.
        If None, the default Blosc2 parameters are used.
    storage : blosc2.Storage or None, optional
        Storage properties for the internal embed store.
        If None, the default Blosc2 storage properties are used.
    threshold : int or None, optional
        Threshold (in bytes of uncompressed data) under which values are kept
        in the embedded store. Default is 0, meaning all values are persisted
        as external files by default.  C2Array objects are always stored in
        the embedded store regardless of this setting.

    Examples
    --------
    >>> dstore = DictStore(localpath="my_dstore.b2z", mode="w")
    >>> dstore["/node1"] = np.array([1, 2, 3])
    >>> dstore["/node2"] = blosc2.ones(2)
    >>> arr_external = blosc2.arange(3, urlpath="ext_node3.b2nd", mode="w")
    >>> dstore["/dir1/node3"] = arr_external  # external file in dir1 (.b2nd)
    >>> schunk = blosc2.SChunk(chunksize=32)
    >>> schunk.append_data(b"abcd")
    4
    >>> dstore["/dir1/schunk1"] = schunk  # externalized as .b2f if above threshold
    >>> dstore.to_b2z(filename="my_dstore.b2z")  # persist to the zip file; external files are copied in
    >>> print(sorted(dstore.keys()))
    ['/dir1/node3', '/dir1/schunk1', '/node1', '/node2']
    >>> print(dstore["/node1"][:]))
    array([1, 2, 3])

    Notes
    -----
    - External persistence uses the following file extensions:
      .b2nd for NDArray, .b2f for SChunk, and .b2b for BatchArray.
      These suffixes are a naming convention for newly written leaves; when
      reopening an existing store, leaf typing is resolved from object
      metadata instead of trusting the suffix alone.
    """

    def __init__(
        self,
        localpath: os.PathLike[Any] | str | bytes,
        mode: str = "a",
        tmpdir: str | None = None,
        cparams: blosc2.CParams | None = None,
        dparams: blosc2.DParams | None = None,
        storage: blosc2.Storage | None = None,
        threshold: int | None = 0,
        *,
        mmap_mode: str | None = None,
        _storage_meta: dict | None = None,
    ):
        """
        See :class:`DictStore` for full documentation of parameters.
        """
        self.localpath = localpath if isinstance(localpath, str | bytes) else str(localpath)
        if mode not in ("r", "w", "a"):
            raise ValueError("For DictStore containers, mode must be 'r', 'w', or 'a'")
        if mmap_mode not in (None, "r"):
            raise ValueError("For DictStore containers, mmap_mode must be None or 'r'")
        if mmap_mode == "r" and mode != "r":
            raise ValueError("For DictStore containers, mmap_mode='r' requires mode='r'")

        self.mode = mode
        self.mmap_mode = mmap_mode
        self.threshold = threshold
        self.cparams = cparams or blosc2.CParams()
        self.dparams = dparams or blosc2.DParams()
        self.storage = storage or blosc2.Storage()

        if _storage_meta:
            self.storage.meta = _storage_meta
        else:
            # Mark this storage as a b2dict object
            self.storage.meta = {"b2dict": {"version": 1}}

        self.offsets = {}
        self.map_tree = {}
        self._temp_dir_obj = None
        self._closed = False
        self._modified = False

        self._setup_paths_and_dirs(tmpdir)

        if self.mode == "r":
            self._init_read_mode(self.dparams)
        else:
            self._init_write_append_mode(self.cparams, self.dparams, storage)

    def _setup_paths_and_dirs(self, tmpdir: str | None):
        """Set up working directories and paths."""
        localpath_exists = os.path.exists(self.localpath)
        if localpath_exists:
            self.is_zip_store = os.path.isfile(self.localpath)
        elif self.localpath.endswith(".b2z"):
            self.is_zip_store = True
        elif self.localpath.endswith(".b2d"):
            self.is_zip_store = False
        else:
            # Default extensionless new stores to directory-backed layout.
            self.is_zip_store = False
        if self.is_zip_store:
            if self.mode == "r":
                # Read mode only needs working_dir as a path namespace for relative↔absolute
                # key arithmetic.  No files are ever written, so no real directory is needed.
                self._temp_dir_obj = None
                self.working_dir = os.path.splitext(os.path.abspath(self.localpath))[0]
            elif tmpdir is None:
                b2z_parent = os.path.dirname(os.path.abspath(self.localpath))
                self._temp_dir_obj = tempfile.TemporaryDirectory(dir=b2z_parent)
                self.working_dir = self._temp_dir_obj.name
            else:
                self._temp_dir_obj = None
                self.working_dir = tmpdir
                os.makedirs(tmpdir, exist_ok=True)
            self.b2z_path = self.localpath
        else:
            self.working_dir = self.localpath
            if self.mode in ("w", "a"):
                os.makedirs(self.working_dir, exist_ok=True)
            if self.localpath.endswith(".b2d"):
                self.b2z_path = self.localpath[:-4] + ".b2z"
            else:
                self.b2z_path = self.localpath + ".b2z"

        self.estore_path = os.path.join(self.working_dir, "embed.b2e")

    def _init_read_mode(self, dparams: blosc2.DParams | None = None):
        """Initialize store in read mode."""
        if not os.path.exists(self.localpath):
            raise FileNotFoundError(f"dir/zip file {self.localpath} does not exist.")

        if self.is_zip_store:
            self.offsets = self._get_zip_offsets()
            if "embed.b2e" not in self.offsets:
                raise FileNotFoundError("Embed file embed.b2e not found in store.")
            estore_offset = self.offsets["embed.b2e"]["offset"]
            schunk = blosc2.blosc2_ext.open(
                self.b2z_path,
                mode="r",
                offset=estore_offset,
                mmap_mode=self.mmap_mode,
                dparams=dparams,
            )
            self._update_map_tree_from_offsets()
        else:  # directory-backed store
            if not os.path.isdir(self.localpath):
                raise FileNotFoundError(f"Directory {self.localpath} does not exist for reading.")
            schunk = blosc2.blosc2_ext.open(
                self.estore_path,
                mode="r",
                offset=0,
                mmap_mode=self.mmap_mode,
                dparams=dparams,
            )
            self._update_map_tree()

        self._estore = EmbedStore(_from_schunk=schunk)
        self.storage.meta = self._estore.storage.meta

    @staticmethod
    def _logical_key_from_relpath(rel_path: str) -> str:
        """Map an external leaf path to its logical tree key."""
        rel_path = rel_path.replace(os.sep, "/")
        key = os.path.splitext(rel_path)[0]
        if not key.startswith("/"):
            key = "/" + key
        return key

    @staticmethod
    def _expected_ext_from_kind(kind: str) -> str:
        """Return the canonical write-time suffix for a supported external leaf kind."""
        if kind == "ndarray":
            return ".b2nd"
        if kind in ("batcharray", "listarray"):
            return ".b2b"
        return ".b2f"

    @classmethod
    def _opened_external_kind(
        cls,
        opened: blosc2.NDArray | SChunk | blosc2.ObjectArray | blosc2.BatchArray | C2Array | Any,
        rel_path: str,
    ) -> str | None:
        """Return the supported external leaf kind for an already opened object."""
        meta = getattr(opened, "schunk", opened).meta
        if "b2o" in meta and isinstance(opened, blosc2.NDArray):
            # Keep b2o carriers as NDArray external leaves during discovery.
            # Rehydrating them here can recurse when a lazy recipe points back
            # into the same DictStore via dictstore_key refs.
            kind = "ndarray"
            processed_name = type(opened).__name__
        else:
            processed = process_opened_object(opened)
            processed_name = type(processed).__name__
            if isinstance(processed, blosc2.BatchArray):
                kind = "batcharray"
            elif isinstance(processed, blosc2.ObjectArray):
                kind = "vlarray"
            elif isinstance(processed, blosc2.NDArray):
                kind = "ndarray"
            elif isinstance(processed, SChunk):
                kind = "schunk"
            elif processed_name == "ListArray":
                kind = "listarray"
            else:
                warnings.warn(
                    f"Ignoring unsupported Blosc2 object at '{rel_path}' during DictStore discovery: "
                    f"{processed_name}",
                    UserWarning,
                    stacklevel=2,
                )
                return None

        expected_ext = cls._expected_ext_from_kind(kind)
        found_ext = os.path.splitext(rel_path)[1]
        if found_ext != expected_ext:
            warnings.warn(
                f"External leaf '{rel_path}' uses extension '{found_ext}' but metadata resolves to "
                f"{processed_name}; expected '{expected_ext}'.",
                UserWarning,
                stacklevel=2,
            )
        return kind

    def _probe_external_leaf_path(self, rel_path: str) -> bool:
        """Return whether a working-dir file is a supported external leaf."""
        urlpath = os.path.join(self.working_dir, rel_path)
        try:
            opened = blosc2.blosc2_ext.open(
                urlpath,
                mode="r",
                offset=0,
                mmap_mode=self.mmap_mode,
                dparams=self.dparams,
            )
        except Exception:
            return False
        return self._opened_external_kind(opened, rel_path) is not None

    def _probe_external_leaf_offset(self, filepath: str) -> bool:
        """Return whether a zip member is a supported external leaf."""
        offset = self.offsets[filepath]["offset"]
        try:
            opened = blosc2.blosc2_ext.open(
                self.b2z_path,
                mode="r",
                offset=offset,
                mmap_mode=self.mmap_mode,
                dparams=self.dparams,
            )
        except Exception:
            return False
        return self._opened_external_kind(opened, filepath) is not None

    def _init_write_append_mode(
        self,
        cparams: blosc2.CParams | None,
        dparams: blosc2.DParams | None,
        storage: blosc2.Storage | None,
    ):
        """Initialize store in write/append mode."""
        if self.mode == "a" and os.path.exists(self.localpath):
            if self.is_zip_store:
                # When using an explicit tmpdir the directory may already contain
                # stale files from a previous open that was never closed.  Clear
                # it before extracting so we always start from a clean slate.
                if self._temp_dir_obj is None:
                    shutil.rmtree(self.working_dir, ignore_errors=True)
                    os.makedirs(self.working_dir, exist_ok=True)
                with zipfile.ZipFile(self.localpath, "r") as zf:
                    zf.extractall(self.working_dir)
            elif not os.path.isdir(self.working_dir):
                raise FileNotFoundError(f"Directory {self.working_dir} does not exist for reading.")

        self._estore = EmbedStore(
            urlpath=self.estore_path,
            mode=self.mode,
            cparams=cparams,
            dparams=dparams,
            storage=storage,
            meta=self.storage.meta,
        )
        self._update_map_tree()

    def _update_map_tree(self):
        """Build map_tree from supported external leaves in working dir.

        Trust canonical external leaf suffixes on the fast path.  Fall back to
        metadata probing for legacy or manually renamed leaves with unusual
        suffixes, preserving discovery warnings and compatibility.
        """
        external_exts = {".b2nd", ".b2f", ".b2b"}
        for root, _, files in os.walk(self.working_dir):
            for file in files:
                filepath = os.path.join(root, file)
                if os.path.abspath(filepath) == os.path.abspath(self.estore_path):
                    continue
                rel_path = os.path.relpath(filepath, self.working_dir).replace(os.sep, "/")
                if os.path.splitext(rel_path)[1] in external_exts or self._probe_external_leaf_path(
                    rel_path
                ):
                    self.map_tree[self._logical_key_from_relpath(rel_path)] = rel_path

    def _update_map_tree_from_offsets(self):
        """Build map_tree from supported external leaves in a zip store.

        Zip-backed stores written by DictStore/TreeStore use canonical external
        leaf suffixes.  Trusting those suffixes avoids opening every member just
        to classify it, which is especially important for compact CTable stores
        with many columns.
        """
        external_exts = {".b2nd", ".b2f", ".b2b"}
        for filepath in self.offsets:
            if filepath == "embed.b2e":
                continue
            if os.path.splitext(filepath)[1] in external_exts or self._probe_external_leaf_offset(filepath):
                self.map_tree[self._logical_key_from_relpath(filepath)] = filepath

    def _annotate_external_value(
        self,
        key: str,
        value: blosc2.NDArray | SChunk | blosc2.ObjectArray | blosc2.BatchArray | C2Array,
    ):
        """Attach DictStore origin metadata so structured msgpack can preserve member identity."""
        value._blosc2_ref = blosc2.Ref.dictstore_key(self.localpath, key)
        return value

    @property
    def estore(self) -> EmbedStore:
        """Access the underlying EmbedStore."""
        return self._estore

    @staticmethod
    def _value_nbytes(value: blosc2.Array | SChunk | blosc2.ObjectArray | blosc2.BatchArray) -> int:
        if isinstance(value, blosc2.ObjectArray | blosc2.BatchArray):
            return value.schunk.nbytes
        return value.nbytes

    @staticmethod
    def _is_external_value(value: blosc2.Array | SChunk | blosc2.ObjectArray | blosc2.BatchArray) -> bool:
        return isinstance(value, blosc2.NDArray | SChunk | blosc2.ObjectArray | blosc2.BatchArray) and bool(
            getattr(value, "urlpath", None)
        )

    @staticmethod
    def _external_ext(value: blosc2.Array | SChunk | blosc2.ObjectArray | blosc2.BatchArray) -> str:
        if isinstance(value, blosc2.NDArray):
            return ".b2nd"
        if isinstance(value, blosc2.BatchArray):
            return ".b2b"
        return ".b2f"

    def __setitem__(
        self, key: str, value: blosc2.Array | SChunk | blosc2.ObjectArray | blosc2.BatchArray
    ) -> None:
        """Add a node to the DictStore."""
        self._modified = True
        if isinstance(value, np.ndarray):
            value = blosc2.asarray(value, cparams=self.cparams, dparams=self.dparams)
        # C2Array should always go to embed store; let estore handle it directly
        if isinstance(value, C2Array):
            self._estore[key] = value
            return
        exceeds_threshold = self.threshold is not None and self._value_nbytes(value) >= self.threshold
        external_file = self._is_external_value(value)
        if exceeds_threshold or (external_file and self.threshold is None):
            ext = self._external_ext(value)
            # Convert key to a proper file path within the tree directory
            rel_key = key.lstrip("/")
            dest_path = os.path.join(self.working_dir, rel_key + ext)

            # Ensure the parent directory exists
            parent_dir = os.path.dirname(dest_path)
            if parent_dir and not os.path.exists(parent_dir):
                os.makedirs(parent_dir, exist_ok=True)

            # Save the value to the destination path
            if not external_file:
                if isinstance(value, blosc2.NDArray) and "b2o" in value.schunk.meta:
                    carrier = blosc2.empty(
                        value.shape,
                        value.dtype,
                        chunks=value.chunks,
                        blocks=value.blocks,
                        cparams=value.cparams,
                        urlpath=dest_path,
                        mode="w",
                        meta={"b2o": value.schunk.meta["b2o"]},
                    )
                    for meta_key, meta_value in value.schunk.vlmeta[:].items():
                        carrier.schunk.vlmeta[meta_key] = meta_value
                elif hasattr(value, "save"):
                    value.save(urlpath=dest_path)
                else:
                    # SChunk, ObjectArray and BatchArray can all be persisted via their cframe.
                    with open(dest_path, "wb") as f:
                        f.write(value.to_cframe())
            else:
                # This should be faster than using value.save() ?
                shutil.copy2(value.urlpath, dest_path)

            # Store relative path from tree directory
            rel_path = os.path.relpath(dest_path, self.working_dir)
            # Normalize to forward slashes
            rel_path = rel_path.replace(os.sep, "/")
            self.map_tree[key] = rel_path
        else:
            # Remove any old external file so it doesn't shadow the embed-stored
            # value on read (map_tree is checked first in __getitem__).
            if key in self.map_tree:
                old_filepath = self.map_tree.pop(key)
                old_full_path = os.path.join(self.working_dir, old_filepath)
                if os.path.exists(old_full_path):
                    os.remove(old_full_path)
            if external_file:
                # Embed a copy by using cframe
                value = blosc2.from_cframe(value.to_cframe())
            self._estore[key] = value

    def __getitem__(
        self, key: str
    ) -> blosc2.NDArray | SChunk | blosc2.ObjectArray | blosc2.BatchArray | C2Array:
        """Retrieve a node from the DictStore."""
        # Check map_tree first
        if key in self.map_tree:
            filepath = self.map_tree[key]
            if filepath in self.offsets:
                offset = self.offsets[filepath]["offset"]
                opened = blosc2.blosc2_ext.open(
                    self.b2z_path,
                    mode="r",
                    offset=offset,
                    mmap_mode=self.mmap_mode,
                    dparams=self.dparams,
                )
                return self._annotate_external_value(key, process_opened_object(opened))
            else:
                urlpath = os.path.join(self.working_dir, filepath)
                if os.path.exists(urlpath):
                    return self._annotate_external_value(
                        key,
                        blosc2.open(
                            urlpath,
                            mode="r" if self.mode == "r" else "a",
                            mmap_mode=self.mmap_mode if self.mode == "r" else None,
                            dparams=self.dparams,
                        ),
                    )
                else:
                    raise KeyError(f"File for key '{key}' not found in offsets or temporary directory.")

        # Fall back to EmbedStore
        return self._estore[key]

    def get(
        self, key: str, default: Any = None
    ) -> blosc2.NDArray | SChunk | blosc2.ObjectArray | blosc2.BatchArray | C2Array | Any:
        """Retrieve a node, or default if not found."""
        try:
            return self[key]
        except KeyError:
            return default

    def __delitem__(self, key: str) -> None:
        """Remove a node from the DictStore."""
        self._modified = True
        if key in self.map_tree:
            # Remove from map_tree and delete the external file
            filepath = self.map_tree[key]
            del self.map_tree[key]

            # Delete the physical file if it exists
            full_path = os.path.join(self.working_dir, filepath)
            if os.path.exists(full_path):
                os.remove(full_path)
        elif key in self._estore:
            del self._estore[key]
        else:
            raise KeyError(f"Key '{key}' not found")

    def __contains__(self, key: str) -> bool:
        """Check if a key exists."""
        return key in self.map_tree or key in self._estore

    def __len__(self) -> int:
        """Return number of nodes."""
        return len(self.map_tree) + len(self._estore)

    def __iter__(self) -> Iterator[str]:
        """Iterate over keys."""
        yield from self.map_tree.keys()
        for key in self._estore:
            if key not in self.map_tree:
                yield key

    def keys(self) -> Set[str]:
        """Return all keys."""
        return self.map_tree.keys() | self._estore.keys()

    def values(self) -> Iterator[blosc2.NDArray | SChunk | C2Array]:
        """Iterate over all values."""
        # Get all unique keys from both map_tree and _estore, with map_tree taking precedence
        all_keys = set(self.map_tree.keys()) | set(self._estore.keys())

        for key in all_keys:
            if key in self.map_tree:
                filepath = self.map_tree[key]
                if self.is_zip_store:
                    if filepath in self.offsets:
                        offset = self.offsets[filepath]["offset"]
                        yield self._annotate_external_value(
                            key,
                            process_opened_object(
                                blosc2.blosc2_ext.open(
                                    self.b2z_path,
                                    mode="r",
                                    offset=offset,
                                    mmap_mode=self.mmap_mode,
                                    dparams=self.dparams,
                                )
                            ),
                        )
                else:
                    urlpath = os.path.join(self.working_dir, filepath)
                    yield self._annotate_external_value(
                        key,
                        blosc2.open(
                            urlpath,
                            mode="r" if self.mode == "r" else "a",
                            mmap_mode=self.mmap_mode if self.mode == "r" else None,
                            dparams=self.dparams,
                        ),
                    )
            elif key in self._estore:
                yield self._estore[key]

    def items(self) -> Iterator[tuple[str, blosc2.NDArray | SChunk | C2Array]]:
        """Iterate over (key, value) pairs."""
        # Get all unique keys from both map_tree and _estore, with map_tree taking precedence
        all_keys = set(self.map_tree.keys()) | set(self._estore.keys())

        for key in all_keys:
            # Check map_tree first, then fall back to _estore
            if key in self.map_tree:
                filepath = self.map_tree[key]
                if self.is_zip_store:
                    if filepath in self.offsets:
                        offset = self.offsets[filepath]["offset"]
                        yield (
                            key,
                            self._annotate_external_value(
                                key,
                                process_opened_object(
                                    blosc2.blosc2_ext.open(
                                        self.b2z_path,
                                        mode="r",
                                        offset=offset,
                                        mmap_mode=self.mmap_mode,
                                        dparams=self.dparams,
                                    )
                                ),
                            ),
                        )
                else:
                    urlpath = os.path.join(self.working_dir, filepath)
                    yield (
                        key,
                        self._annotate_external_value(
                            key,
                            blosc2.open(
                                urlpath,
                                mode="r" if self.mode == "r" else "a",
                                mmap_mode=self.mmap_mode if self.mode == "r" else None,
                                dparams=self.dparams,
                            ),
                        ),
                    )
            elif key in self._estore:
                yield key, self._estore[key]

    def to_b2z(self, overwrite=False, filename=None) -> os.PathLike[Any] | str:
        """
        Serialize store contents to a compact ``.b2z`` file.

        Parameters
        ----------
        overwrite : bool, optional
            If True, overwrite the existing b2z file if it exists. Default is False.
        filename : str, optional
            If provided, use this filename instead of the default b2z file path.
            Keyword use is recommended for clarity.

        Returns
        -------
        filename : str
            The absolute path to the created b2z file.

        Examples
        --------
        Pack a directory-backed store into a zip store.  A ``.b2d`` suffix is
        recommended for directory-backed stores, but not required::

            with blosc2.DictStore("data.b2d", mode="w") as dstore:
                dstore["/values"] = np.arange(10)

            with blosc2.DictStore("data.b2d", mode="r") as dstore:
                dstore.to_b2z(filename="data.b2z", overwrite=True)

        ``filename`` can also be passed positionally::

            with blosc2.DictStore("data.b2d", mode="r") as dstore:
                dstore.to_b2z("copy.b2z", overwrite=True)
        """
        if isinstance(overwrite, str | os.PathLike) and filename is None:
            filename = overwrite
            overwrite = False

        if self.mode == "r" and self.is_zip_store:
            raise ValueError("Cannot call to_b2z() on a .b2z DictStore opened in read mode.")

        b2z_path = self.b2z_path if filename is None else filename
        b2z_path = os.fspath(b2z_path)
        if not b2z_path.endswith(".b2z"):
            raise ValueError("b2z_path must have a .b2z extension")

        if os.path.exists(b2z_path) and not overwrite:
            raise FileExistsError(f"'{b2z_path}' already exists. Use overwrite=True to overwrite.")

        # Gather all files except estore_path
        filepaths = []
        for root, _, files in os.walk(self.working_dir):
            for file in files:
                filepath = os.path.join(root, file)
                if os.path.abspath(filepath) != os.path.abspath(self.estore_path):
                    filepaths.append(filepath)

        # Sort filepaths by file size from largest to smallest
        filepaths.sort(key=os.path.getsize, reverse=True)

        with zipfile.ZipFile(b2z_path, "w", zipfile.ZIP_STORED) as zf:
            # Write all files (except estore_path) first (sorted by size)
            for filepath in filepaths:
                arcname = os.path.relpath(filepath, self.working_dir)
                zf.write(filepath, arcname)
            # Write estore last
            if os.path.exists(self.estore_path):
                arcname = os.path.relpath(self.estore_path, self.working_dir)
                zf.write(self.estore_path, arcname)
        return os.path.abspath(b2z_path)

    def to_b2d(self, dirname=None, *, overwrite: bool = False) -> os.PathLike[Any] | str:
        """
        Serialize store contents to a directory-backed store.

        Parameters
        ----------
        dirname : str, optional
            If provided, use this directory instead of the default directory
            path.  A ``.b2d`` suffix is recommended for clarity, but not
            required.
        overwrite : bool, optional
            If True, overwrite the existing b2d directory if it exists.
            Default is False.

        Returns
        -------
        dirname : str
            The absolute path to the created directory-backed store.

        Examples
        --------
        Unpack a zip-backed store into a directory-backed store::

            with blosc2.DictStore("data.b2z", mode="r") as dstore:
                dstore.to_b2d("data.b2d", overwrite=True)

            with blosc2.DictStore("data.b2d", mode="r") as dstore:
                values = dstore["/values"][:]

        Copy an existing directory-backed store to another directory.  A
        ``.b2d`` suffix is recommended for directory-backed stores::

            with blosc2.DictStore("data.b2d", mode="r") as dstore:
                dstore.to_b2d("backup.b2d", overwrite=True)
        """
        b2d_path = self.localpath if dirname is None and not self.is_zip_store else dirname
        if b2d_path is None:
            b2d_path = (
                self.b2z_path[:-4] + ".b2d" if self.b2z_path.endswith(".b2z") else self.b2z_path + ".b2d"
            )
        b2d_path = os.fspath(b2d_path)

        target_path = os.path.abspath(b2d_path)
        source_path = os.path.abspath(self.working_dir)
        if not self.is_zip_store and target_path == source_path:
            return target_path

        if os.path.exists(target_path):
            if not overwrite:
                raise FileExistsError(f"'{target_path}' already exists. Use overwrite=True to overwrite.")
            if os.path.isdir(target_path):
                shutil.rmtree(target_path)
            else:
                os.remove(target_path)

        if self.is_zip_store and self.mode == "r":
            os.makedirs(target_path, exist_ok=True)
            with zipfile.ZipFile(self.b2z_path, "r") as zf:
                zf.extractall(target_path)
        else:
            shutil.copytree(self.working_dir, target_path)
        return target_path

    def _get_zip_offsets(self) -> dict[str, dict[str, int]]:
        """Get offset and length of all files in the zip archive."""
        self.offsets = {}  # Reset offsets
        with open(self.b2z_path, "rb") as f, zipfile.ZipFile(f) as zf:
            for info in zf.infolist():
                # info.header_offset points to the local file header
                # The actual file data starts after the header
                f.seek(info.header_offset)
                local_header = f.read(30)
                filename_len = int.from_bytes(local_header[26:28], "little")
                extra_len = int.from_bytes(local_header[28:30], "little")
                data_offset = info.header_offset + 30 + filename_len + extra_len
                self.offsets[info.filename] = {"offset": data_offset, "length": info.file_size}
        return self.offsets

    def close(self) -> None:
        """Persist changes and cleanup."""
        if self._closed:
            return
        self._closed = True

        # Repack estore
        # TODO: for some reason this is not working
        # if self.mode != "r":
        #     cframe = self._estore.to_cframe()
        #     with open(self._estore.urlpath, "wb") as f:
        #         f.write(cframe)

        if self.is_zip_store and self.mode in ("w", "a"):
            # Serialize to b2z file.
            self.to_b2z(overwrite=True)

        # Clean up temporary directory if we created it
        if self._temp_dir_obj is not None:
            self._temp_dir_obj.cleanup()

    def discard(self) -> None:
        """Clean up resources *without* repacking the .b2z file.

        Use this instead of :meth:`close` when the store was opened only for
        inspection and should be thrown away without persisting any changes
        back to the archive.
        """
        if self._closed:
            return
        self._closed = True
        if self._temp_dir_obj is not None:
            self._temp_dir_obj.cleanup()

    def __del__(self):
        """Ensure the temporary directory is removed and, if writes were made
        through this store's own API, the store is flushed back to the .b2z
        file.

        When no Python-level writes went through ``__setitem__`` / ``__delitem__``
        (``_modified`` is False), we skip ``to_b2z()`` to avoid repacking a
        potentially partial temp dir during garbage collection.  Explicit
        ``close()`` / ``__exit__`` always repacks regardless.
        """
        try:
            if not self._closed and self.is_zip_store and self.mode in ("w", "a") and not self._modified:
                # Skip repacking — discard is safe and avoids corrupting the
                # archive when the temp dir is torn down during GC.
                self.discard()
            else:
                self.close()
        except Exception:
            pass

    def __enter__(self):
        """Context manager enter."""
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """Context manager exit."""
        self.close()
        # No need to handle exceptions, just close the DictStore
        return False


if __name__ == "__main__":
    # Example usage
    localpath = "example_dstore.b2z"
    if True:
        with DictStore(localpath, mode="w") as dstore:
            dstore["/node1"] = np.array([1, 2, 3])
            dstore["/node2"] = blosc2.ones(2)

            # Make /node3 an external file
            arr_external = blosc2.arange(3, urlpath="ext_node3.b2nd", mode="w")
            dstore["/dir1/node3"] = arr_external

            print("DictStore keys:", list(dstore.keys()))
            print("Node1 data:", dstore["/node1"][:])
            print("Node2 data:", dstore["/node2"][:])
            print("Node3 data (external):", dstore["/dir1/node3"][:])

            del dstore["/node1"]
            print("After deletion, keys:", list(dstore.keys()))

    # Open the stored zip file
    with DictStore(localpath, mode="r") as dstore_opened:
        print("Opened dstore keys:", list(dstore_opened.keys()))
        for key, value in dstore_opened.items():
            if isinstance(value, blosc2.NDArray):
                print(
                    f"Key: {key}, Shape: {value.shape}, Values: {value[:10] if len(value) > 3 else value[:]}"
                )



#######################################################################
# Copyright (c) 2019-present, Blosc Development Team <blosc@blosc.org>
# All rights reserved.
#
# SPDX-License-Identifier: BSD-3-Clause
#######################################################################

from __future__ import annotations

import ast
import contextlib
import inspect
import os
import textwrap
import tokenize
from io import StringIO
from typing import ClassVar

_PRINT_DSL_KERNEL = os.environ.get("PRINT_DSL_KERNEL", "").strip().lower()
_PRINT_DSL_KERNEL = _PRINT_DSL_KERNEL not in ("", "0", "false", "no", "off")
_DSL_USAGE_DOC_URL = "https://github.com/Blosc/python-blosc2/blob/main/doc/getting_started/dsl_syntax.md"


class DSLSyntaxError(ValueError):
    """Raised when a @dsl_kernel function uses unsupported DSL syntax."""


def _normalize_miniexpr_scalar(value):
    # NumPy scalar-like values expose .item(); plain Python scalars do not.
    # Do not call .item() on non-scalar arrays; for Blosc2 arrays this can be expensive.
    if hasattr(value, "shape") and value.shape != ():
        raise TypeError("Unsupported scalar type for miniexpr specialization")
    if hasattr(value, "item") and callable(value.item):
        with contextlib.suppress(Exception):
            value = value.item()
    if isinstance(value, bool):
        return int(value)
    if isinstance(value, int | float | str):
        return value
    raise TypeError("Unsupported scalar type for miniexpr specialization")


def _line_starts(text: str) -> list[int]:
    starts = [0]
    for i, ch in enumerate(text):
        if ch == "\n":
            starts.append(i + 1)
    return starts


def _to_abs(line_starts: list[int], line: int, col: int) -> int:
    return line_starts[line - 1] + col


def _find_def_signature_span(text: str):
    tokens = list(tokenize.generate_tokens(StringIO(text).readline))
    for i, tok in enumerate(tokens):
        if tok.type != tokenize.NAME or tok.string != "def":
            continue
        lparen = None
        rparen = None
        colon = None
        depth = 0
        for j in range(i + 1, len(tokens)):
            t = tokens[j]
            if lparen is None:
                if t.type == tokenize.OP and t.string == "(":
                    lparen = t
                    depth = 1
                continue
            if t.type == tokenize.OP and t.string == "(":
                depth += 1
                continue
            if t.type == tokenize.OP and t.string == ")":
                depth -= 1
                if depth == 0:
                    rparen = t
                    continue
            if rparen is not None and t.type == tokenize.OP and t.string == ":":
                colon = t
                break
        if lparen is not None and rparen is not None:
            return lparen, rparen, colon
    return None, None, None


def _remove_scalar_params_preserving_source(text: str, scalar_replacements: dict[str, int | float]):
    if not scalar_replacements:
        return text, 0

    lparen, rparen, colon = _find_def_signature_span(text)
    if lparen is None or rparen is None:
        return text, 0

    try:
        tree = ast.parse(text)
    except Exception:
        return text, 0

    func = next((n for n in tree.body if isinstance(n, ast.FunctionDef)), None)
    if func is None:
        return text, 0

    kept = [a.arg for a in (func.args.posonlyargs + func.args.args) if a.arg not in scalar_replacements]
    line_starts = _line_starts(text)
    pstart = _to_abs(line_starts, lparen.end[0], lparen.end[1])
    pend = _to_abs(line_starts, rparen.start[0], rparen.start[1])
    updated = f"{text[:pstart]}{', '.join(kept)}{text[pend:]}"
    body_start = 0
    if colon is not None:
        # Signature shrink can move ':' to an earlier column, so recompute
        # on the rewritten text to avoid skipping first-line body tokens.
        _, _, updated_colon = _find_def_signature_span(updated)
        if updated_colon is not None:
            body_start = _to_abs(_line_starts(updated), updated_colon.end[0], updated_colon.end[1])
    return updated, body_start


def _replace_scalar_names_preserving_source(
    text: str, scalar_replacements: dict[str, int | float], body_start: int
):
    if not scalar_replacements:
        return text

    line_starts = _line_starts(text)
    tokens = list(tokenize.generate_tokens(StringIO(text).readline))
    significant = {
        tokenize.NAME,
        tokenize.NUMBER,
        tokenize.STRING,
        tokenize.OP,
        tokenize.INDENT,
        tokenize.DEDENT,
    }
    assign_ops = {"=", "+=", "-=", "*=", "/=", "//=", "%=", "&=", "|=", "^=", "<<=", ">>=", ":="}
    edits = []
    for i, tok in enumerate(tokens):
        if tok.type != tokenize.NAME or tok.string not in scalar_replacements:
            continue
        start_abs = _to_abs(line_starts, tok.start[0], tok.start[1])
        if start_abs < body_start:
            continue

        prev_sig = None
        for j in range(i - 1, -1, -1):
            if tokens[j].type in significant:
                prev_sig = tokens[j]
                break
        if prev_sig is not None and prev_sig.type == tokenize.OP and prev_sig.string == ".":
            continue

        next_sig = None
        for j in range(i + 1, len(tokens)):
            if tokens[j].type in significant:
                next_sig = tokens[j]
                break
        if next_sig is not None and next_sig.type == tokenize.OP and next_sig.string in assign_ops:
            continue

        end_abs = _to_abs(line_starts, tok.end[0], tok.end[1])
        edits.append((start_abs, end_abs, repr(scalar_replacements[tok.string])))

    if not edits:
        return text

    out = text
    for start, end, repl in sorted(edits, key=lambda e: e[0], reverse=True):
        out = f"{out[:start]}{repl}{out[end:]}"
    return out


def _fold_numeric_cast_calls_preserving_source(text: str, body_start: int):
    """Fold float(<number>) and int(<number>) calls into literals.

    miniexpr parses DSL function calls in a restricted way, and scalar specialization can
    produce calls like float(200) that fail to parse.  Fold those into literals while
    preserving source formatting/comments elsewhere.
    """
    try:
        tree = ast.parse(text)
    except Exception:
        return text

    line_starts = _line_starts(text)
    edits = []

    def _numeric_literal_value(node):
        if isinstance(node, ast.Constant) and isinstance(node.value, int | float | bool):
            return node.value
        if (
            isinstance(node, ast.UnaryOp)
            and isinstance(node.op, ast.UAdd | ast.USub)
            and isinstance(node.operand, ast.Constant)
            and isinstance(node.operand.value, int | float | bool)
        ):
            value = node.operand.value
            return value if isinstance(node.op, ast.UAdd) else -value
        return None

    for node in ast.walk(tree):
        if not isinstance(node, ast.Call):
            continue
        if node.keywords or len(node.args) != 1:
            continue
        if not isinstance(node.func, ast.Name) or node.func.id not in {"float", "int"}:
            continue

        arg = node.args[0]
        value = _numeric_literal_value(arg)
        if value is None:
            continue

        start_abs = _to_abs(line_starts, node.lineno, node.col_offset)
        if start_abs < body_start:
            continue
        end_abs = _to_abs(line_starts, node.end_lineno, node.end_col_offset)

        if node.func.id == "float":
            repl = repr(float(value))
        else:
            repl = repr(int(value))
        edits.append((start_abs, end_abs, repl))

    if not edits:
        return text

    out = text
    for start, end, repl in sorted(edits, key=lambda e: e[0], reverse=True):
        out = f"{out[:start]}{repl}{out[end:]}"
    return out


def specialize_miniexpr_inputs(expr_string: str, operands: dict):
    """Inline scalar operands as constants for miniexpr compilation."""
    scalar_replacements = {}
    array_operands = {}
    for name, value in operands.items():
        if hasattr(value, "shape") and value.shape == ():
            scalar_replacements[name] = _normalize_miniexpr_scalar(value[()])
            continue
        if isinstance(value, int | float | bool | str) or (
            not hasattr(value, "shape") and hasattr(value, "item") and callable(value.item)
        ):
            try:
                scalar_replacements[name] = _normalize_miniexpr_scalar(value)
                continue
            except TypeError:
                pass
        array_operands[name] = value

    if not scalar_replacements:
        return expr_string, operands

    rewritten, body_start = _remove_scalar_params_preserving_source(expr_string, scalar_replacements)
    rewritten = _replace_scalar_names_preserving_source(rewritten, scalar_replacements, body_start)
    rewritten = _fold_numeric_cast_calls_preserving_source(rewritten, body_start)
    return rewritten, array_operands


def specialize_dsl_miniexpr_inputs(expr_string: str, operands: dict):
    """Backward-compatible alias for DSL-specific callers."""
    return specialize_miniexpr_inputs(expr_string, operands)


class DSLValidator:
    _binop_map: ClassVar[dict[type[ast.operator], str]] = {
        ast.Add: "+",
        ast.Sub: "-",
        ast.Mult: "*",
        ast.Div: "/",
        ast.FloorDiv: "//",
        ast.Mod: "%",
        ast.Pow: "**",
        ast.BitAnd: "&",
        ast.BitOr: "|",
        ast.BitXor: "^",
        ast.LShift: "<<",
        ast.RShift: ">>",
    }
    _cmp_map: ClassVar[dict[type[ast.cmpop], str]] = {
        ast.Eq: "==",
        ast.NotEq: "!=",
        ast.Lt: "<",
        ast.LtE: "<=",
        ast.Gt: ">",
        ast.GtE: ">=",
    }

    def __init__(self, source: str, line_base: int = 0, input_names: list[str] | None = None):
        self._source = source
        self._line_base = line_base
        self._inputs = set(input_names or ())

    def validate(self, func_node: ast.FunctionDef):
        self._args(func_node)
        if not func_node.body:
            self._err(func_node, "DSL kernel must have a body")
        self._one_per_line(func_node.body)
        for stmt in func_node.body:
            self._stmt(stmt)

    def _one_per_line(self, body: list[ast.stmt]):
        # G1: miniexpr parses one statement per line; `;`-joined siblings share a lineno.
        prev = None
        for stmt in body:
            if prev is not None and stmt.lineno == prev:
                self._err(
                    stmt,
                    "Only one statement per line is supported in DSL kernels; "
                    "split ';'-joined statements onto separate lines",
                )
            prev = stmt.lineno

    def _err(self, node: ast.AST, msg: str, *, line: int | None = None, col: int | None = None):
        if line is None:
            line = getattr(node, "lineno", 0)
        if col is None:
            col = getattr(node, "col_offset", 0) + 1
        line -= self._line_base
        location = f"{msg} at line {line}, column {col}"
        dump = self._format_source_with_pointer(line, col)
        raise DSLSyntaxError(f"{location}\n\nDSL kernel source:\n{dump}\n\nSee: {_DSL_USAGE_DOC_URL}")

    def _format_source_with_pointer(self, line: int, col: int) -> str:
        lines = self._source.splitlines()
        if not lines:
            return "<empty>"
        width = len(str(len(lines)))
        out = []
        for lineno, text in enumerate(lines, start=1):
            out.append(f"{lineno:>{width}} | {text}")
            if lineno == line:
                pointer = " " * max(col - 1, 0)
                out.append(f"{' ' * width} | {pointer}^")
        return "\n".join(out)

    def _args(self, func_node: ast.FunctionDef):
        args = func_node.args
        if args.vararg or args.kwarg or args.kwonlyargs:
            self._err(args, "DSL kernel does not support *args/**kwargs/kwonly args")
        if args.defaults or args.kw_defaults:
            self._err(args, "DSL kernel does not support default arguments")

    def _check_input_assign(self, target: ast.Name):
        # G2: miniexpr forbids reassigning an input parameter (inputs alias operand buffers).
        if target.id in self._inputs:
            self._err(
                target,
                f"Cannot assign to input parameter '{target.id}'; "
                f"copy it into a local variable first (e.g. 'tmp = {target.id}')",
            )

    def _stmt(self, node: ast.stmt):  # noqa: C901
        if isinstance(node, ast.Assign):
            if len(node.targets) != 1 or not isinstance(node.targets[0], ast.Name):
                self._err(node, "Only simple assignments are supported in DSL kernels")
            self._check_input_assign(node.targets[0])
            self._expr(node.value)
            return
        if isinstance(node, ast.AugAssign):
            if not isinstance(node.target, ast.Name):
                self._err(node, "Only simple augmented assignments are supported")
            self._check_input_assign(node.target)
            self._binop(node.op)
            self._expr(node.value)
            return
        if isinstance(node, ast.Return):
            if node.value is None:
                self._err(node, "DSL kernel return must have a value")
            self._expr(node.value)
            return
        if isinstance(node, ast.Expr):
            self._expr(node.value)
            return
        if isinstance(node, ast.If):
            self._expr(node.test)
            if not node.body:
                self._err(node, "Empty if blocks are not supported in DSL kernels")
            self._one_per_line(node.body)
            self._one_per_line(node.orelse)
            for stmt in node.body:
                self._stmt(stmt)
            for stmt in node.orelse:
                self._stmt(stmt)
            return
        if isinstance(node, ast.For):
            if node.orelse:
                self._err(node, "for/else is not supported in DSL kernels")
            if not isinstance(node.target, ast.Name):
                self._err(node, "DSL for-loop target must be a simple name")
            if not isinstance(node.iter, ast.Call):
                self._err(node, "DSL for-loop must iterate over range()")
            func_name = self._call_name(node.iter.func)
            if func_name != "range":
                self._err(node, "DSL for-loop must iterate over range()")
            if node.iter.keywords or not (1 <= len(node.iter.args) <= 3):
                self._err(node, "DSL range() must take 1 to 3 positional arguments")
            for arg in node.iter.args:
                self._expr(arg)
            if not node.body:
                self._err(node, "Empty for-loop bodies are not supported in DSL kernels")
            self._one_per_line(node.body)
            for stmt in node.body:
                self._stmt(stmt)
            return
        if isinstance(node, ast.While):
            if node.orelse:
                self._err(node, "while/else is not supported in DSL kernels")
            self._expr(node.test)
            if not node.body:
                self._err(node, "Empty while-loop bodies are not supported in DSL kernels")
            self._one_per_line(node.body)
            for stmt in node.body:
                self._stmt(stmt)
            return
        if isinstance(node, ast.Break | ast.Continue):
            return
        self._err(node, f"Unsupported DSL statement: {type(node).__name__}")

    def _expr(self, node: ast.AST):  # noqa: C901
        if isinstance(node, ast.Name):
            return
        if isinstance(node, ast.Constant):
            val = node.value
            if isinstance(val, bool | int | float | str):
                return
            self._err(node, "Unsupported constant in DSL expression")
        if isinstance(node, ast.UnaryOp):
            if isinstance(node.op, ast.UAdd | ast.USub | ast.Not):
                self._expr(node.operand)
                return
            self._err(node, "Unsupported unary operator in DSL expression")
        if isinstance(node, ast.BinOp):
            self._binop(node.op)
            self._expr(node.left)
            self._expr(node.right)
            return
        if isinstance(node, ast.BoolOp):
            for value in node.values:
                self._expr(value)
            return
        if isinstance(node, ast.Compare):
            if len(node.ops) != 1 or len(node.comparators) != 1:
                self._err(node, "Chained comparisons are not supported in DSL")
            self._cmpop(node.ops[0])
            self._expr(node.left)
            self._expr(node.comparators[0])
            return
        if isinstance(node, ast.Call):
            self._call_name(node.func)
            if node.keywords:
                self._err(node, "Keyword arguments are not supported in DSL calls")
            for arg in node.args:
                self._expr(arg)
            return
        if isinstance(node, ast.IfExp):
            seg = ast.get_source_segment(self._source, node)
            col = getattr(node, "col_offset", 0) + 1
            if seg is not None:
                rel = seg.find(" if ")
                if rel >= 0:
                    col += rel + 1
            self._err(
                node,
                "Ternary expressions are not supported in DSL; use where(cond, a, b)",
                col=col,
            )
        self._err(node, f"Unsupported DSL expression: {type(node).__name__}")

    def _call_name(self, node: ast.AST) -> str:
        if isinstance(node, ast.Name):
            return node.id
        if (
            isinstance(node, ast.Attribute)
            and isinstance(node.value, ast.Name)
            and node.value.id in {"np", "numpy", "math"}
        ):
            return node.attr
        self._err(node, "Unsupported call target in DSL")
        raise AssertionError("unreachable")

    def _binop(self, op: ast.operator):
        for k in self._binop_map:
            if isinstance(op, k):
                return
        self._err(op, "Unsupported binary operator in DSL")

    def _cmpop(self, op: ast.cmpop):
        for k in self._cmp_map:
            if isinstance(op, k):
                return
        self._err(op, "Unsupported comparison in DSL")


class DSLKernel:
    """Wrap a Python function and optionally extract a miniexpr DSL kernel from it."""

    def __init__(self, func):
        self.func = func
        self.__name__ = getattr(func, "__name__", self.__class__.__name__)
        self.__qualname__ = getattr(func, "__qualname__", self.__name__)
        self.__doc__ = getattr(func, "__doc__", None)
        try:
            sig = inspect.signature(func)
        except (TypeError, ValueError):
            sig = None
        self._sig = sig
        self._sig_has_varargs = False
        self._sig_npositional = None
        self._legacy_udf_signature = False
        if sig is not None:
            params = list(sig.parameters.values())
            positional_params = [p for p in params if p.kind in (p.POSITIONAL_ONLY, p.POSITIONAL_OR_KEYWORD)]
            self._sig_has_varargs = any(p.kind == p.VAR_POSITIONAL for p in params)
            self._sig_npositional = len(positional_params)
            # Preserve support for classic lazyudf signature: (inputs_tuple, output, offset)
            if not self._sig_has_varargs and len(positional_params) == 3:
                p2 = positional_params[1].name.lower()
                p3 = positional_params[2].name.lower()
                self._legacy_udf_signature = p2 in {"output", "out"} and p3 == "offset"
        self.dsl_source = None
        self.input_names = None
        self.dsl_error = None
        try:
            dsl_source, input_names = self._extract_dsl(func)
        except DSLSyntaxError as e:
            # Preserve extracted source/signature for diagnostics even when DSL validation fails.
            try:
                dsl_source, input_names = self._extract_dsl(func, validate=False)
            except Exception:
                dsl_source = None
                input_names = None
            self.dsl_error = e
        except Exception:
            dsl_source = None
            input_names = None
        self.dsl_source = dsl_source
        self.input_names = input_names

    def _extract_dsl(self, func, validate: bool = True):
        source = inspect.getsource(func)
        source = textwrap.dedent(source)
        tree = ast.parse(source)
        func_node = None
        for node in tree.body:
            if isinstance(node, ast.FunctionDef) and node.name == func.__name__:
                func_node = node
                break
        if func_node is None:
            for node in tree.body:
                if isinstance(node, ast.FunctionDef):
                    func_node = node
                    break
        if func_node is None:
            raise ValueError("No function definition found for DSL extraction")

        dsl_source = self._slice_function_source(source, func_node)
        dsl_tree = ast.parse(dsl_source)
        dsl_func = next((node for node in dsl_tree.body if isinstance(node, ast.FunctionDef)), None)
        if dsl_func is None:
            raise ValueError("No function definition found in sliced DSL source")
        input_names = self._input_names_from_signature(dsl_func)
        if validate:
            DSLValidator(dsl_source, input_names=input_names).validate(dsl_func)
        if _PRINT_DSL_KERNEL:
            func_name = getattr(func, "__name__", "<dsl_kernel>")
            print(f"[DSLKernel:{func_name}] dsl_source (full):")
            print(dsl_source)
        return dsl_source, input_names

    @staticmethod
    def _slice_function_source(source: str, func_node: ast.FunctionDef) -> str:
        lines = source.splitlines()
        start = func_node.lineno - 1
        end_lineno = getattr(func_node, "end_lineno", None)
        if end_lineno is None:
            end = len(lines)
        else:
            end = end_lineno
        return "\n".join(lines[start:end])

    @staticmethod
    def _input_names_from_signature(func_node: ast.FunctionDef) -> list[str]:
        args = func_node.args
        if args.vararg or args.kwarg or args.kwonlyargs:
            raise ValueError("DSL kernel does not support *args/**kwargs/kwonly args")
        if args.defaults or args.kw_defaults:
            raise ValueError("DSL kernel does not support default arguments")
        return [a.arg for a in (args.posonlyargs + args.args)]

    def __call__(self, inputs_tuple, output, offset=None):
        if self.dsl_error is not None:
            raise self.dsl_error
        if self._legacy_udf_signature:
            return self.func(inputs_tuple, output, offset)

        n_inputs = len(inputs_tuple)
        if self._sig is not None and (
            self._sig_npositional in (n_inputs, n_inputs + 1) or self._sig_has_varargs
        ):
            if self._sig_npositional == n_inputs + 1:
                result = self.func(*inputs_tuple, offset)
            else:
                result = self.func(*inputs_tuple)
            output[...] = result
            return None

        try:
            return self.func(inputs_tuple, output, offset)
        except TypeError:
            result = self.func(*inputs_tuple)
            output[...] = result
            return None


def dsl_kernel(func):
    """Decorator to wrap a function in a DSLKernel."""

    return DSLKernel(func)


def validate_dsl(func):
    """Validate a DSL kernel function without executing it.

    Parameters
    ----------
    func
        A Python callable or :class:`DSLKernel`.

    Returns
    -------
    dict
        A dictionary with:
        - ``valid`` (bool): whether the DSL is valid
        - ``dsl_source`` (str | None): extracted DSL source when valid
        - ``input_names`` (list[str] | None): input signature names when valid
        - ``error`` (str | None): user-facing error message when invalid

    Examples
    --------
    >>> import blosc2
    >>> @blosc2.dsl_kernel
    ... def k(a, b):
    ...     return a * a + b * b
    >>> info = blosc2.validate_dsl(k)
    >>> info["valid"]
    True
    >>> info["input_names"]
    ['a', 'b']

    An unsupported construct is reported instead of raising:

    >>> @blosc2.dsl_kernel
    ... def bad(a):
    ...     return 1 if a > 0 else 0
    >>> blosc2.validate_dsl(bad)["valid"]
    False

    See Also
    --------
    validate_dsl_jit : Additionally probe whether the kernel JIT-compiles (vs falls
        back to the interpreter) for given operand/output dtypes.
    """

    kernel = func if isinstance(func, DSLKernel) else DSLKernel(func)
    err = kernel.dsl_error
    return {
        "valid": err is None,
        "dsl_source": kernel.dsl_source,
        "input_names": kernel.input_names,
        "error": None if err is None else str(err),
    }


def validate_dsl_jit(func, operands, out_dtype, *, shape=(64,), chunks=None, blocks=None):
    """Report whether a DSL kernel JIT-compiles (vs. interpreter fallback).

    Compiles a tiny probe of the kernel for the given operand/output dtypes and
    queries miniexpr for whether it produced a runtime JIT kernel.  The kernel is
    *not* run on real data, but compilation is dtype-specialized, so the operand and
    output dtypes must be supplied.

    Parameters
    ----------
    func
        A Python callable or :class:`DSLKernel`.
    operands
        One spec per kernel input.  A NumPy dtype (e.g. ``np.float64``, ``"f8"``)
        marks an *array* operand; a Python ``int``/``float``/``bool`` marks a *scalar*
        parameter inlined as a constant (as happens at run time).  Either a dict
        mapping input name -> spec, or a sequence matched positionally to the inputs.
    out_dtype
        Output dtype to specialize the codegen for.
    shape
        Probe shape; ``len(shape)`` sets the kernel dimensionality (kernels using
        ``_i1`` etc. need a matching ndim).  Defaults to ``(64,)``.

    Returns
    -------
    dict
        - ``valid`` (bool): DSL syntax is valid
        - ``jit`` (bool): a runtime JIT kernel was produced (vs interpreter fallback)
        - ``compiled`` (bool): miniexpr accepted the kernel
        - ``status`` (str | None): miniexpr compile-status name
        - ``error`` (str | None): syntax error message when ``valid`` is False

    Examples
    --------
    >>> import blosc2
    >>> import numpy as np
    >>> @blosc2.dsl_kernel
    ... def k(a, b):
    ...     return a * a + b * b
    >>> status = blosc2.validate_dsl_jit(k, [np.float64, np.float64], np.float64)
    >>> status["jit"]
    True
    >>> status["status"]
    'ME_COMPILE_SUCCESS'

    Pass array operands as dtypes and scalar parameters as values (the scalar is
    inlined as a constant, as at run time):

    >>> @blosc2.dsl_kernel
    ... def scaled(a, n):
    ...     return a * float(n)
    >>> blosc2.validate_dsl_jit(scaled, {"a": np.float64, "n": 3}, np.float64)["jit"]
    True

    See Also
    --------
    validate_dsl : Static DSL-syntax check only (no compilation; no dtypes needed).
    """
    import numpy as np

    import blosc2

    kernel = func if isinstance(func, DSLKernel) else DSLKernel(func)
    if kernel.dsl_error is not None:
        return {
            "valid": False,
            "jit": False,
            "compiled": False,
            "status": None,
            "error": str(kernel.dsl_error),
        }

    names = list(kernel.input_names or [])
    if isinstance(operands, dict):
        specs = dict(operands)
    else:
        operands = list(operands)
        if len(operands) != len(names):
            raise ValueError(f"expected {len(names)} operand specs for inputs {names}, got {len(operands)}")
        specs = dict(zip(names, operands, strict=True))

    scalars: dict = {}
    arrays: dict = {}
    for name in names:
        if name not in specs:
            raise ValueError(f"missing operand spec for input '{name}'")
        spec = specs[name]
        if isinstance(spec, bool | int | float):
            scalars[name] = spec
        else:
            arrays[name] = np.dtype(spec)

    source = kernel.dsl_source
    if scalars:
        source, _ = specialize_miniexpr_inputs(source, scalars)
    if not arrays:
        raise ValueError(
            "validate_dsl_jit needs at least one array operand (pass a dtype spec); "
            "all-scalar probes are not supported"
        )

    out = blosc2.empty(shape, dtype=np.dtype(out_dtype), chunks=chunks, blocks=blocks)
    probe_inputs = {name: np.empty(1, dtype=dt) for name, dt in arrays.items()}
    status = out._dsl_jit_status(source, probe_inputs)
    status["valid"] = True
    status["error"] = None
    return status


def kernel_from_source(source: str, name: str | None = None) -> DSLKernel:
    """Reconstruct a :class:`DSLKernel` from its stored source text.

    Executes *source* in a restricted namespace (builtins minus ``__import__``,
    plus ``np`` and ``blosc2``), extracts the defined function, and wraps it in
    a :class:`DSLKernel`.  This is the inverse of persisting
    :attr:`DSLKernel.dsl_source` and is shared by the persisted-``LazyUDF``
    decoder and the CTable DSL-column loaders.

    Parameters
    ----------
    source
        Complete, standalone function-definition source (as produced by
        :attr:`DSLKernel.dsl_source` or :func:`inspect.getsource`).
    name
        Name of the function to extract from *source*.  When omitted, the
        single top-level function definition in *source* is used.
    """
    import builtins
    import linecache

    import numpy as np

    import blosc2

    tree = ast.parse(source)
    func_defs = [n for n in tree.body if isinstance(n, ast.FunctionDef)]
    if name is None:
        if len(func_defs) != 1:
            raise ValueError(
                "kernel_from_source requires an explicit 'name' when 'source' does not "
                "contain exactly one top-level function definition"
            )
        name = func_defs[0].name

    # Security: reject sources that include top-level statements with
    # side effects (assignments, calls, loops, etc.) beyond the function
    # definition itself.  Imports and docstrings are permitted.
    allowed_nodes = (ast.FunctionDef, ast.Import, ast.ImportFrom)
    for node in tree.body:
        if isinstance(node, allowed_nodes):
            continue
        # A bare string literal at module level is a docstring — safe.
        if (
            isinstance(node, ast.Expr)
            and isinstance(node.value, ast.Constant)
            and isinstance(node.value.value, str)
        ):
            continue
        raise ValueError(
            f"kernel_from_source source must contain only a function definition "
            f"(optionally preceded by imports), but found: {ast.dump(node)[:120]}"
        )

    local_ns: dict = {}
    filename = f"<{name}>"
    safe_globals = {
        "__builtins__": {k: v for k, v in builtins.__dict__.items() if k != "__import__"},
        "np": np,
        "blosc2": blosc2,
    }
    linecache.cache[filename] = (len(source), None, source.splitlines(True), filename)
    exec(compile(source, filename, "exec"), safe_globals, local_ns)
    try:
        func = local_ns[name]
    except KeyError:
        raise ValueError(f"kernel_from_source: source did not define function {name!r}") from None
    if not isinstance(func, DSLKernel):
        func = DSLKernel(func)
    return func


class DSLBuilder:
    _binop_map: ClassVar[dict[type[ast.operator], str]] = {
        ast.Add: "+",
        ast.Sub: "-",
        ast.Mult: "*",
        ast.Div: "/",
        ast.FloorDiv: "//",
        ast.Mod: "%",
        ast.Pow: "**",
        ast.BitAnd: "&",
        ast.BitOr: "|",
        ast.BitXor: "^",
        ast.LShift: "<<",
        ast.RShift: ">>",
    }

    _cmp_map: ClassVar[dict[type[ast.cmpop], str]] = {
        ast.Eq: "==",
        ast.NotEq: "!=",
        ast.Lt: "<",
        ast.LtE: "<=",
        ast.Gt: ">",
        ast.GtE: ">=",
    }

    def __init__(self):
        self._lines = []

    def build(self, func_node: ast.FunctionDef):
        input_names = self._args(func_node.args)
        self._emit(f"def {func_node.name}({', '.join(input_names)}):", 0)
        if not func_node.body:
            raise ValueError("DSL kernel must have a body")
        for stmt in func_node.body:
            self._stmt(stmt, 4)
        return "\n".join(self._lines), input_names

    def _emit(self, line: str, indent: int):
        self._lines.append(" " * indent + line)

    def _args(self, args: ast.arguments):
        if args.vararg or args.kwarg or args.kwonlyargs:
            raise ValueError("DSL kernel does not support *args/**kwargs/kwonly args")
        if args.defaults or args.kw_defaults:
            raise ValueError("DSL kernel does not support default arguments")
        names = [a.arg for a in (args.posonlyargs + args.args)]
        if not names:
            raise ValueError("DSL kernel must accept at least one argument")
        return names

    def _stmt(self, node: ast.stmt, indent: int):
        if isinstance(node, ast.Assign):
            if len(node.targets) != 1 or not isinstance(node.targets[0], ast.Name):
                raise ValueError("Only simple assignments are supported in DSL kernels")
            target = node.targets[0].id
            value = self._expr(node.value)
            self._emit(f"{target} = {value}", indent)
            return
        if isinstance(node, ast.AugAssign):
            if not isinstance(node.target, ast.Name):
                raise ValueError("Only simple augmented assignments are supported")
            target = node.target.id
            op = self._binop(node.op)
            value = self._expr(node.value)
            self._emit(f"{target} = {target} {op} {value}", indent)
            return
        if isinstance(node, ast.Return):
            if node.value is None:
                raise ValueError("DSL kernel return must have a value")
            value = self._expr(node.value)
            self._emit(f"return {value}", indent)
            return
        if isinstance(node, ast.Expr):
            value = self._expr(node.value)
            self._emit(value, indent)
            return
        if isinstance(node, ast.If):
            self._if_stmt(node, indent)
            return
        if isinstance(node, ast.For):
            self._for_stmt(node, indent)
            return
        if isinstance(node, ast.While):
            self._while_stmt(node, indent)
            return
        if isinstance(node, ast.Break):
            self._emit("break", indent)
            return
        if isinstance(node, ast.Continue):
            self._emit("continue", indent)
            return
        raise ValueError(f"Unsupported DSL statement: {type(node).__name__}")

    def _stmt_block(self, body, indent: int):
        if not body:
            raise ValueError("Empty blocks are not supported in DSL kernels")
        i = 0
        while i < len(body):
            stmt = body[i]
            if (
                isinstance(stmt, ast.If)
                and not stmt.orelse
                and self._block_terminates(stmt.body)
                and i + 1 < len(body)
                and isinstance(body[i + 1], ast.If)
            ):
                merged = ast.If(test=stmt.test, body=stmt.body, orelse=[body[i + 1]])
                self._if_stmt(merged, indent)
                i += 2
                continue
            self._stmt(stmt, indent)
            i += 1

    def _block_terminates(self, body) -> bool:
        if not body:
            return False
        return self._stmt_terminates(body[-1])

    def _stmt_terminates(self, node: ast.stmt) -> bool:
        if isinstance(node, ast.Return | ast.Break | ast.Continue):
            return True
        if isinstance(node, ast.If) and node.orelse:
            return self._block_terminates(node.body) and self._block_terminates(node.orelse)
        return False

    def _if_stmt(self, node: ast.If, indent: int):
        current = node
        first = True
        while True:
            prefix = "if" if first else "elif"
            cond = self._expr(current.test)
            self._emit(f"{prefix} {cond}:", indent)
            self._stmt_block(current.body, indent + 4)
            first = False
            if current.orelse and len(current.orelse) == 1 and isinstance(current.orelse[0], ast.If):
                current = current.orelse[0]
                continue
            break
        if current.orelse:
            self._emit("else:", indent)
            self._stmt_block(current.orelse, indent + 4)

    def _for_stmt(self, node: ast.For, indent: int):
        if node.orelse:
            raise ValueError("for/else is not supported in DSL kernels")
        if not isinstance(node.target, ast.Name):
            raise ValueError("DSL for-loop target must be a simple name")
        if not isinstance(node.iter, ast.Call):
            raise ValueError("DSL for-loop must iterate over range()")
        func_name = self._call_name(node.iter.func)
        if func_name != "range":
            raise ValueError("DSL for-loop must iterate over range()")
        if node.iter.keywords or len(node.iter.args) != 1:
            raise ValueError("DSL range() must take a single argument")
        limit = self._expr(node.iter.args[0])
        self._emit(f"for {node.target.id} in range({limit}):", indent)
        self._stmt_block(node.body, indent + 4)

    def _while_stmt(self, node: ast.While, indent: int):
        if node.orelse:
            raise ValueError("while/else is not supported in DSL kernels")
        cond = self._expr(node.test)
        self._emit(f"while {cond}:", indent)
        self._stmt_block(node.body, indent + 4)

    def _expr(self, node: ast.AST) -> str:  # noqa: C901
        if isinstance(node, ast.Name):
            return node.id
        if isinstance(node, ast.Constant):
            val = node.value
            if isinstance(val, bool):
                return "1" if val else "0"
            if isinstance(val, int | float):
                return repr(val)
            raise ValueError("Unsupported constant in DSL expression")
        if isinstance(node, ast.UnaryOp):
            if isinstance(node.op, ast.UAdd):
                return f"+{self._expr(node.operand)}"
            if isinstance(node.op, ast.USub):
                return f"-{self._expr(node.operand)}"
            if isinstance(node.op, ast.Not):
                return f"!{self._expr(node.operand)}"
            raise ValueError("Unsupported unary operator in DSL expression")
        if isinstance(node, ast.BinOp):
            left = self._expr(node.left)
            right = self._expr(node.right)
            op = self._binop(node.op)
            return f"({left} {op} {right})"
        if isinstance(node, ast.BoolOp):
            op = "&" if isinstance(node.op, ast.And) else "|"
            values = [self._expr(v) for v in node.values]
            expr = values[0]
            for val in values[1:]:
                expr = f"({expr} {op} {val})"
            return expr
        if isinstance(node, ast.Compare):
            if len(node.ops) != 1 or len(node.comparators) != 1:
                raise ValueError("Chained comparisons are not supported in DSL")
            left = self._expr(node.left)
            right = self._expr(node.comparators[0])
            op = self._cmpop(node.ops[0])
            return f"({left} {op} {right})"
        if isinstance(node, ast.Call):
            func_name = self._call_name(node.func)
            if node.keywords:
                raise ValueError("Keyword arguments are not supported in DSL calls")
            args = ", ".join(self._expr(a) for a in node.args)
            return f"{func_name}({args})"
        if isinstance(node, ast.IfExp):
            cond = self._expr(node.test)
            body = self._expr(node.body)
            orelse = self._expr(node.orelse)
            return f"where({cond}, {body}, {orelse})"
        raise ValueError(f"Unsupported DSL expression: {type(node).__name__}")

    def _call_name(self, node: ast.AST) -> str:
        if isinstance(node, ast.Name):
            return node.id
        if (
            isinstance(node, ast.Attribute)
            and isinstance(node.value, ast.Name)
            and node.value.id in {"np", "numpy", "math"}
        ):
            return node.attr
        raise ValueError("Unsupported call target in DSL")

    def _binop(self, op: ast.operator) -> str:
        for k, v in self._binop_map.items():
            if isinstance(op, k):
                return v
        raise ValueError("Unsupported binary operator in DSL")

    def _cmpop(self, op: ast.cmpop) -> str:
        for k, v in self._cmp_map.items():
            if isinstance(op, k):
                return v
        raise ValueError("Unsupported comparison in DSL")


class DSLReducer:
    _binop_map: ClassVar[dict[type[ast.operator], str]] = DSLBuilder._binop_map
    _cmp_map: ClassVar[dict[type[ast.cmpop], str]] = DSLBuilder._cmp_map

    def __init__(self, max_unroll: int = 64):
        self._env: dict[str, str] = {}
        self._const_env: dict[str, object] = {}
        self._return_expr: str | None = None
        self._max_unroll = max_unroll

    def reduce(self, func_node: ast.FunctionDef):
        input_names = self._args(func_node.args)
        if not func_node.body:
            return None
        for stmt in func_node.body:
            if not self._stmt(stmt):
                return None
            if self._return_expr is not None:
                break
        if self._return_expr is None:
            return None
        return self._return_expr, input_names

    def _args(self, args: ast.arguments):
        if args.vararg or args.kwarg or args.kwonlyargs:
            raise ValueError("DSL kernel does not support *args/**kwargs/kwonly args")
        if args.defaults or args.kw_defaults:
            raise ValueError("DSL kernel does not support default arguments")
        names = [a.arg for a in (args.posonlyargs + args.args)]
        if not names:
            raise ValueError("DSL kernel must accept at least one argument")
        return names

    def _stmt(self, node: ast.stmt) -> bool:  # noqa: C901
        if isinstance(node, ast.Assign):
            if len(node.targets) != 1 or not isinstance(node.targets[0], ast.Name):
                return False
            target = node.targets[0].id
            value = self._expr(node.value)
            self._env[target] = value
            const_val = self._const_eval(node.value)
            if const_val is None:
                self._const_env.pop(target, None)
            else:
                self._const_env[target] = const_val
            return True
        if isinstance(node, ast.AugAssign):
            if not isinstance(node.target, ast.Name):
                return False
            target = node.target.id
            op = self._binop(node.op)
            value = self._expr(node.value)
            left = self._env.get(target, target)
            left_const = self._const_env.get(target)
            right_const = self._const_eval(node.value)
            simplified = self._simplify_binop_expr(op, left, value, left_const, right_const)
            self._env[target] = simplified
            if left_const is None or right_const is None:
                self._const_env.pop(target, None)
            else:
                self._const_env[target] = self._apply_binop(left_const, right_const, node.op)
            return True
        if isinstance(node, ast.Return):
            if node.value is None:
                return False
            self._return_expr = self._expr(node.value)
            return True
        if isinstance(node, ast.If):
            test_val = self._const_eval(node.test)
            if test_val is None:
                return False
            branch = node.body if bool(test_val) else node.orelse
            if not branch:
                return True
            for stmt in branch:
                if not self._stmt(stmt):
                    return False
                if self._return_expr is not None:
                    return True
            return True
        if isinstance(node, ast.For):
            if node.orelse:
                return False
            if not isinstance(node.target, ast.Name):
                return False
            if not isinstance(node.iter, ast.Call):
                return False
            func_name = self._call_name(node.iter.func)
            if func_name != "range":
                return False
            if node.iter.keywords or len(node.iter.args) != 1:
                return False
            limit_val = self._const_eval(node.iter.args[0])
            if limit_val is None or not isinstance(limit_val, int):
                return False
            if limit_val < 0 or limit_val > self._max_unroll:
                return False
            loop_var = node.target.id
            old_env = self._env.get(loop_var)
            old_const = self._const_env.get(loop_var)
            for i in range(limit_val):
                self._env[loop_var] = str(i)
                self._const_env[loop_var] = i
                for stmt in node.body:
                    if not self._stmt(stmt):
                        if old_env is None:
                            self._env.pop(loop_var, None)
                        else:
                            self._env[loop_var] = old_env
                        if old_const is None:
                            self._const_env.pop(loop_var, None)
                        else:
                            self._const_env[loop_var] = old_const
                        return False
                    if self._return_expr is not None:
                        break
                if self._return_expr is not None:
                    break
            if old_env is None:
                self._env.pop(loop_var, None)
            else:
                self._env[loop_var] = old_env
            if old_const is None:
                self._const_env.pop(loop_var, None)
            else:
                self._const_env[loop_var] = old_const
            return True
        return False

    def _expr(self, node: ast.AST) -> str:  # noqa: C901
        const_val = self._const_eval(node)
        if const_val is not None:
            if isinstance(const_val, bool):
                return "1" if const_val else "0"
            return repr(const_val)
        if isinstance(node, ast.Name):
            if node.id in self._env:
                val = self._env[node.id]
                # Avoid double-wrapping if already parenthesized or is a function call
                if (val.startswith("(") and val.endswith(")")) or "(" in val:
                    return val
                return f"({val})"
            return node.id
        if isinstance(node, ast.Constant):
            val = node.value
            if isinstance(val, bool):
                return "1" if val else "0"
            if isinstance(val, int | float):
                return repr(val)
            raise ValueError("Unsupported constant in DSL expression")
        if isinstance(node, ast.UnaryOp):
            if isinstance(node.op, ast.UAdd):
                return f"+{self._expr(node.operand)}"
            if isinstance(node.op, ast.USub):
                return f"-{self._expr(node.operand)}"
            if isinstance(node.op, ast.Not):
                return f"!{self._expr(node.operand)}"
            raise ValueError("Unsupported unary operator in DSL expression")
        if isinstance(node, ast.BinOp):
            left = self._expr(node.left)
            right = self._expr(node.right)
            op = self._binop(node.op)
            left_const = self._const_eval(node.left)
            right_const = self._const_eval(node.right)
            return self._simplify_binop_expr(op, left, right, left_const, right_const)
        if isinstance(node, ast.BoolOp):
            op = "&" if isinstance(node.op, ast.And) else "|"
            values = [self._expr(v) for v in node.values]
            expr = values[0]
            for val in values[1:]:
                expr = f"({expr} {op} {val})"
            return expr
        if isinstance(node, ast.Compare):
            if len(node.ops) != 1 or len(node.comparators) != 1:
                raise ValueError("Chained comparisons are not supported in DSL")
            left = self._expr(node.left)
            right = self._expr(node.comparators[0])
            op = self._cmpop(node.ops[0])
            return f"({left} {op} {right})"
        if isinstance(node, ast.Call):
            func_name = self._call_name(node.func)
            if node.keywords:
                raise ValueError("Keyword arguments are not supported in DSL calls")
            args = ", ".join(self._expr(a) for a in node.args)
            return f"{func_name}({args})"
        if isinstance(node, ast.IfExp):
            cond = self._expr(node.test)
            body = self._expr(node.body)
            orelse = self._expr(node.orelse)
            return f"where({cond}, {body}, {orelse})"
        raise ValueError(f"Unsupported DSL expression: {type(node).__name__}")

    def _call_name(self, node: ast.AST) -> str:
        if isinstance(node, ast.Name):
            return node.id
        if (
            isinstance(node, ast.Attribute)
            and isinstance(node.value, ast.Name)
            and node.value.id in {"np", "numpy", "math"}
        ):
            return node.attr
        raise ValueError("Unsupported call target in DSL")

    def _binop(self, op: ast.operator) -> str:
        for k, v in self._binop_map.items():
            if isinstance(op, k):
                return v
        raise ValueError("Unsupported binary operator in DSL")

    def _cmpop(self, op: ast.cmpop) -> str:
        for k, v in self._cmp_map.items():
            if isinstance(op, k):
                return v
        raise ValueError("Unsupported comparison in DSL")

    def _const_eval(self, node: ast.AST):  # noqa: C901
        if isinstance(node, ast.Constant):
            if isinstance(node.value, int | float | bool):
                return node.value
            return None
        if isinstance(node, ast.Name):
            return self._const_env.get(node.id)
        if isinstance(node, ast.UnaryOp):
            val = self._const_eval(node.operand)
            if val is None:
                return None
            if isinstance(node.op, ast.UAdd):
                return +val
            if isinstance(node.op, ast.USub):
                return -val
            if isinstance(node.op, ast.Not):
                return not val
            return None
        if isinstance(node, ast.BinOp):
            left = self._const_eval(node.left)
            right = self._const_eval(node.right)
            if left is None or right is None:
                return None
            return self._apply_binop(left, right, node.op)
        if isinstance(node, ast.BoolOp):
            vals = [self._const_eval(v) for v in node.values]
            if any(v is None for v in vals):
                return None
            if isinstance(node.op, ast.And):
                return all(vals)
            if isinstance(node.op, ast.Or):
                return any(vals)
            return None
        if isinstance(node, ast.Compare):
            if len(node.ops) != 1 or len(node.comparators) != 1:
                return None
            left = self._const_eval(node.left)
            right = self._const_eval(node.comparators[0])
            if left is None or right is None:
                return None
            return self._apply_cmp(left, right, node.ops[0])
        return None

    def _apply_binop(self, left, right, op):
        if isinstance(op, ast.Add):
            return left + right
        if isinstance(op, ast.Sub):
            return left - right
        if isinstance(op, ast.Mult):
            return left * right
        if isinstance(op, ast.Div):
            return left / right
        if isinstance(op, ast.FloorDiv):
            return left // right
        if isinstance(op, ast.Mod):
            return left % right
        if isinstance(op, ast.Pow):
            return left**right
        if isinstance(op, ast.BitAnd):
            return left & right
        if isinstance(op, ast.BitOr):
            return left | right
        if isinstance(op, ast.BitXor):
            return left ^ right
        if isinstance(op, ast.LShift):
            return left << right
        if isinstance(op, ast.RShift):
            return left >> right
        return None

    def _apply_cmp(self, left, right, op):
        if isinstance(op, ast.Eq):
            return left == right
        if isinstance(op, ast.NotEq):
            return left != right
        if isinstance(op, ast.Lt):
            return left < right
        if isinstance(op, ast.LtE):
            return left <= right
        if isinstance(op, ast.Gt):
            return left > right
        if isinstance(op, ast.GtE):
            return left >= right
        return None

    def _simplify_binop_expr(self, op, left_expr, right_expr, left_const, right_const):
        if op == "+":
            if self._is_zero(left_const):
                return right_expr
            if self._is_zero(right_const):
                return left_expr
        if op == "-" and self._is_zero(right_const):
            return left_expr
        if op == "*":
            if self._is_one(left_const):
                return right_expr
            if self._is_one(right_const):
                return left_expr
        return f"({left_expr} {op} {right_expr})"

    def _is_zero(self, value):
        return isinstance(value, int | float | bool) and value == 0

    def _is_one(self, value):
        return isinstance(value, int | float | bool) and value == 1
